/****************************************************************************
*
*    The MIT License (MIT)
*
*    Copyright (c) 2014 - 2022 Vivante Corporation
*
*    Permission is hereby granted, free of charge, to any person obtaining a
*    copy of this software and associated documentation files (the "Software"),
*    to deal in the Software without restriction, including without limitation
*    the rights to use, copy, modify, merge, publish, distribute, sublicense,
*    and/or sell copies of the Software, and to permit persons to whom the
*    Software is furnished to do so, subject to the following conditions:
*
*    The above copyright notice and this permission notice shall be included in
*    all copies or substantial portions of the Software.
*
*    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
*    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
*    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
*    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
*    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
*    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
*    DEALINGS IN THE SOFTWARE.
*
*****************************************************************************
*
*    The GPL License (GPL)
*
*    Copyright (C) 2014 - 2022 Vivante Corporation
*
*    This program is free software; you can redistribute it and/or
*    modify it under the terms of the GNU General Public License
*    as published by the Free Software Foundation; either version 2
*    of the License, or (at your option) any later version.
*
*    This program is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with this program; if not, write to the Free Software Foundation,
*    Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
*
*****************************************************************************
*
*    Note: This software is released under dual MIT and GPL licenses. A
*    recipient may use this file under the terms of either the MIT license or
*    GPL License. If you wish to use only one license not the other, you can
*    indicate your decision by deleting one of the above license notices in your
*    version of this file.
*
*****************************************************************************/

#include "gc_hal_kernel_linux.h"
#include "gc_hal_kernel_platform.h"
#include "gc_hal_kernel_platform_supernova.h"
#include "transcoder.h"
#include "hdma.h"

#define _GC_OBJ_ZONE                    gcvZONE_EVENT

/* Disable MSI for internal FPGA build except PPC */
#if gcdFPGA_BUILD
# define USE_MSI            0
#else
# define USE_MSI            1
#endif

#define gcdMIXED_PLATFORM   0

#define gcdDISABLE_NODE_OFFSET 1

gceSTATUS
_AdjustParam(gcsPLATFORM *Platform, gcsMODULE_PARAMETERS *Args);

gceSTATUS
_GetGPUPhysical(gcsPLATFORM *Platform, gctPHYS_ADDR_T CPUPhysical, gctPHYS_ADDR_T *GPUPhysical);

#if gcdENABLE_MP_SWITCH
gceSTATUS
_SwitchCoreCount(gcsPLATFORM *Platform, gctUINT32 *Count);
#endif

#if gcdENABLE_VIDEO_MEMORY_MIRROR
gceSTATUS
_dmaCopy(gctPOINTER Object, gcsDMA_TRANS_INFO *Info);
#endif


static struct _gcsPLATFORM_OPERATIONS default_ops = {
    .adjustParam        = _AdjustParam,
    .getGPUPhysical     = _GetGPUPhysical,
#if gcdENABLE_MP_SWITCH
    .switchCoreCount    = _SwitchCoreCount,
#endif
#if gcdENABLE_VIDEO_MEMORY_MIRROR
    .dmaCopy            = _dmaCopy,
#endif
};

#if 0
struct _gcsPLATFORM_PCIE default_platform = {
    .base = {
        .name = __FILE__,
        .ops  = &default_ops,
    },
};
#endif

gctINT
_QueryBarInfo(struct pci_dev *Pdev, gctPHYS_ADDR_T *BarAddr, gctUINT64 *BarSize, gctUINT BarNum)
{
    gctUINT addr, size;
    gctINT is_64_bit = 0;
    gctUINT64 addr64, size64;

    /* Read the bar address */
    if (pci_read_config_dword(Pdev, PCI_BASE_ADDRESS_0 + BarNum * 0x4, &addr) < 0)
        return -1;

    /* Read the bar size */
    if (pci_write_config_dword(Pdev, PCI_BASE_ADDRESS_0 + BarNum * 0x4, 0xffffffff) < 0)
        return -1;

    if (pci_read_config_dword(Pdev, PCI_BASE_ADDRESS_0 + BarNum * 0x4, &size) < 0)
        return -1;

    /* Write back the bar address */
    if (pci_write_config_dword(Pdev, PCI_BASE_ADDRESS_0 + BarNum * 0x4, addr) < 0)
        return -1;

    /* The bar is not working properly */
    if (size == 0xffffffff)
        return -1;

    if ((size & PCI_BASE_ADDRESS_MEM_TYPE_MASK) == PCI_BASE_ADDRESS_MEM_TYPE_64)
        is_64_bit = 1;

    addr64 = addr;
    size64 = size;

    if (is_64_bit) {
        /* Read the bar address */
        if (pci_read_config_dword(Pdev, PCI_BASE_ADDRESS_0 + (BarNum + 1) * 0x4, &addr) < 0)
            return -1;

        /* Read the bar size */
        if (pci_write_config_dword(Pdev, PCI_BASE_ADDRESS_0 + (BarNum + 1) * 0x4, 0xffffffff) < 0)
            return -1;

        if (pci_read_config_dword(Pdev, PCI_BASE_ADDRESS_0 + (BarNum + 1) * 0x4, &size) < 0)
            return -1;

        /* Write back the bar address */
        if (pci_write_config_dword(Pdev, PCI_BASE_ADDRESS_0 + (BarNum + 1) * 0x4, addr) < 0)
            return -1;

        addr64 |= ((gctUINT64)addr << 32);
        size64 |= ((gctUINT64)size << 32);
    }

    size64 &= ~0xfULL;
    size64 = ~size64;
    size64 += 1;
    addr64 &= ~0xfULL;

    gcmkPRINT("Bar%d addr=0x%08lx size=0x%08lx", BarNum, addr64, size64);

    *BarAddr = addr64;
    *BarSize = size64;

    return is_64_bit;
}

static const struct pci_device_id vivpci_ids[] = {
  {
    .class       = 0x000000,
    .class_mask  = 0x000000,
    .vendor      = 0x10ee,
    .device      = 0x7012,
    .subvendor   = PCI_ANY_ID,
    .subdevice   = PCI_ANY_ID,
    .driver_data = 0
  },
  {
    .class       = 0x000000,
    .class_mask  = 0x000000,
    .vendor      = 0x10ee,
    .device      = 0x7014,
    .subvendor   = PCI_ANY_ID,
    .subdevice   = PCI_ANY_ID,
    .driver_data = 0
  },
  { /* End: all zeroes */ }
};

MODULE_DEVICE_TABLE(pci, vivpci_ids);

#define ENABLE_AXI_SRAM     1  //use default AXI sram

gceSTATUS
_AdjustParam(gcsPLATFORM *Platform, gcsMODULE_PARAMETERS *Args)
{
    struct _gcsPLATFORM_PCIE *pcie_platform = (struct _gcsPLATFORM_PCIE *)Platform;
    unsigned int i;
    unsigned int dev_index;
    struct sn_tranx_t *tdev;
    unsigned long flags = 0;
    gctUINT64 S1_offset, S2_offset;
    static gctINT32 local_sRAMOffsets[SRAM_NUM_PER_DEVICE] = {
    [0 ... SRAM_NUM_PER_DEVICE - 1] = -1,
    [0] = S1_AXI_SRAM_OFF,
    [1] = S2_AXI_SRAM_OFF,
    };

    tdev = (struct sn_tranx_t *)pci_get_drvdata(pcie_platform->pcie_info[0].pdev);

    switch(tdev->pf_vf_mode)
    {
        case PF_MODE:
            S1_offset = THS1_PCIE_INT_EN_CON_STUS;
            S2_offset = THS2_PCIE_INT_EN_CON_STUS;
            break;
        case TWO_VF_MODE:
            if(tdev->vf_index == VF1_INDEX) {
                Args->irqs[2] = -1;
                Args->irqs[3] = -1;
#if S2_GPU
                Args->irq2Ds[1] = -1;
#endif
            }
            else if(tdev->vf_index == VF2_INDEX) {
                Args->irqs[gcvCORE_MAJOR] = -1;
                Args->irqs[1] = -1;
#if S1_GPU
                Args->irq2Ds[0] = -1;
#endif
            }
            S1_offset = VF_THS_PCIE_INT_EN_CON_STUS;
            S2_offset = VF_THS_PCIE_INT_EN_CON_STUS;
            local_sRAMOffsets[0] = local_sRAMOffsets[1] = VF_AXI_SRAM_OFF;
            break;
        case ONE_VF_MODE:
            S1_offset = ONE_VF_THS1_PCIE_INT_EN_CON_STUS;
            S2_offset = ONE_VF_THS2_PCIE_INT_EN_CON_STUS;
            local_sRAMOffsets[0] = ONE_VF_S1_AXI_SRAM_OFF;
            local_sRAMOffsets[1] = ONE_VF_S2_AXI_SRAM_OFF;
            break;
        default:
            sn_pri(tdev, SN_INF, "%s,vf_max_count incorrect\n", __func__);
            break;
    }

    for (dev_index = 0; dev_index < pcie_platform->device_number; dev_index++)
    {
        struct pci_dev *pcieDev = pcie_platform->pcie_info[dev_index].pdev;

        for (i = 0; i < gcdMAX_PCIE_BAR; i++) {
            flags = pci_resource_flags(pcieDev, i);
            flags = (flags & IORESOURCE_MEM_64) ? 1 : 0;
            pcie_platform->pcie_info[dev_index].bar[i].base = pci_resource_start(pcieDev, i);
            pcie_platform->pcie_info[dev_index].bar[i].size = pci_resource_len(pcieDev, i);
            pcie_platform->pcie_info[dev_index].bar[i].available = gcvTRUE;

            pr_info("query bar info %d, %llx, %llx.", i, pcie_platform->pcie_info[dev_index].bar[i].base, pcie_platform->pcie_info[dev_index].bar[i].size);
            if (flags == 0)
                continue;

            i += flags;
        }
        /* limite vip's size */
        pcie_platform->pcie_info[dev_index].bar[4].size = BAR4_VIP_LENGTH;

        pcie_platform->pcie_info[dev_index].bar[2].logical = (gctPOINTER)pci_iomap(pcieDev, 2, pcie_platform->pcie_info[dev_index].bar[2].size);
        if (!pcie_platform->pcie_info[dev_index].bar[2].logical)
            gcmkTRACE_ZONE_N(gcvLEVEL_INFO, gcvZONE_EVENT, 40, "bar[%d] map failed\n",i);

        pr_info("bar base info %llx, %llx, %llx.", pcie_platform->pcie_info[dev_index].bar[0].base, pcie_platform->pcie_info[dev_index].bar[2].base,
               pcie_platform->pcie_info[dev_index].bar[4].base);

        for (i = 0; i < gcvCORE_COUNT; i++)
        {
            if (Args->irqs[i] != -1)
            {
                gcmkTRACE_ZONE(gcvLEVEL_ERROR, gcvZONE_EVENT,"core id = %d\n", i);
                {
                    unsigned int val;
                    if(Args->irqs[i] == 29 || Args->irqs[i] == 30)
                    {
                        val = readl(pcie_platform->pcie_info[dev_index].bar[2].logical + S2_offset);
                        val |= 1 << Args->irqs[i];
                        writel(val,pcie_platform->pcie_info[dev_index].bar[2].logical + S2_offset);
                    }
                    else if (Args->irqs[i] == 19 || Args->irqs[i] == 20)
                    {
                        val = readl(pcie_platform->pcie_info[dev_index].bar[2].logical + S1_offset);
                        val |= 1 << Args->irqs[i];
                        writel(val,pcie_platform->pcie_info[dev_index].bar[2].logical + S1_offset);
                    }
                    else
                    {
                        gcmkTRACE_ZONE(gcvLEVEL_ERROR, gcvZONE_EVENT,"Args->irqs[i] = %d\n", Args->irqs[i]);
                        continue;
                    }
                }

                gcmkTRACE_ZONE(gcvLEVEL_ERROR, gcvZONE_EVENT, "irq number before map: Args->irqs[%d] = %d\n", i, Args->irqs[i]);
                Args->irqs[i] = pci_irq_vector(pcieDev, Args->irqs[i]);
                gcmkTRACE_ZONE(gcvLEVEL_ERROR, gcvZONE_EVENT, "irq number after map: Args->irqs[%d] = %d\n", i, Args->irqs[i]);
            }
        }

        for (i = 0; i < gcdGLOBAL_2D_COUNT; i++) {
            if (Args->irq2Ds[i] != -1) {
                Args->irq2Ds[i] = pci_irq_vector(pcieDev, Args->irq2Ds[i]);
            }
        }

        for (i = 0; i < gcvCORE_COUNT; i++)
        {
            if (i >= gcvCORE_2D && i <= gcvCORE_2D_MAX)
                Args->register2DBasesMapped[i - gcvCORE_2D] =
                    pcie_platform->pcie_info[dev_index].bar[2].logical + Args->regOffsets[i];
            else
                Args->registerBasesMapped[i] =
                    pcie_platform->pcie_info[dev_index].bar[2].logical + Args->regOffsets[i];
        }

#if ENABLE_AXI_SRAM
        /* All the PCIE devices AXI-SRAM should have same base address. */
        /* Get CPU view SRAM base address from bar address and bar inside offset. */
        for (i = 0; i < SRAM_NUM_PER_DEVICE; i++) {
            int sram_bar = -1;
            Args->sRAMOffsets[dev_index][i] = local_sRAMOffsets[i];
            pcie_platform->pcie_info[dev_index].sram_bar[i] = Args->sRAMBars[dev_index][i];
            pcie_platform->pcie_info[dev_index].sram_offset[i] = Args->sRAMOffsets[dev_index][i];
            sram_bar = Args->sRAMBars[dev_index][i];
            gcmkASSERT(pcie_platform->pcie_info[dev_index].bar[sram_bar].available);

            Args->extSRAMBases[dev_index][i] = pcie_platform->pcie_info[dev_index].bar[sram_bar].base +
                                               pcie_platform->pcie_info[dev_index].sram_offset[i];
            Args->extSRAMSizes[dev_index][i] = 0x200000;

            pcie_platform->pcie_info[dev_index].sram_base[i] = Args->extSRAMBases[dev_index][i];
            if (i % SRAM_NUM_PER_DEVICE)
                pcie_platform->pcie_info[dev_index].sram_gpu_base[i] = 0x417000000;
            else
                pcie_platform->pcie_info[dev_index].sram_gpu_base[i] = 0x413000000;
            pcie_platform->pcie_info[dev_index].sram_size[i] = Args->extSRAMSizes[dev_index][i];
            pr_info("axi-sram, bar(%d), base(0x%llx), gpubase(0x%llx), size(0x%x)\n", sram_bar, pcie_platform->pcie_info[dev_index].sram_base[i],
                    pcie_platform->pcie_info[dev_index].sram_gpu_base[i], pcie_platform->pcie_info[dev_index].sram_size[i]);
        }
#endif

        Args->exclusiveBase[dev_index] = 0x7F00000;
        Args->exclusiveSize[dev_index] = 0x100000;

        Args->externalBase[dev_index] = pcie_platform->pcie_info[dev_index].bar[4].base + BAR4_OFFSET_VALUE;
        Args->externalSize[dev_index] = pcie_platform->pcie_info[dev_index].bar[4].size - BAR4_OFFSET_VALUE;

        Args->contiguousBases[dev_index] = gcvINVALID_PHYSICAL_ADDRESS;
        Args->contiguousSizes[dev_index] = 0;
        gcmkTRACE_ZONE(gcvLEVEL_INFO, gcvZONE_DRIVER, "size info is %llx, %llx \n", pcie_platform->pcie_info[dev_index].bar[4].base, pcie_platform->pcie_info[dev_index].bar[4].size);

        dev_index++;
    }

    return gcvSTATUS_OK;
}

gceSTATUS
_GetGPUPhysical(gcsPLATFORM *Platform, gctPHYS_ADDR_T CPUPhysical, gctPHYS_ADDR_T *GPUPhysical)
{
    struct _gcsPLATFORM_PCIE *pcie_platform = (struct _gcsPLATFORM_PCIE *)Platform;
    gctPHYS_ADDR_T external_base;
    int32_t external_size;
    size_t offset = 0;
    unsigned int pdev_index = 0;

    for (pdev_index = 0; pdev_index < pcie_platform->device_number; pdev_index++) {
        external_base = pcie_platform->pcie_info[pdev_index].bar[4].base;
        external_size = pcie_platform->pcie_info[pdev_index].bar[4].size;

        if ((external_size > 0) && (CPUPhysical >= external_base) && (CPUPhysical < (external_base + external_size))) {
            *GPUPhysical = 0x0 + CPUPhysical - external_base + offset;
            return gcvSTATUS_OK;
        }

#if ENABLE_AXI_SRAM
        {
            gctPHYS_ADDR_T extsram_base;
            int32_t extsram_size, i = 0;

            for (i = 0; i < SRAM_NUM_PER_DEVICE; i++) {
                extsram_base = pcie_platform->pcie_info[pdev_index].sram_base[i];
                extsram_size = pcie_platform->pcie_info[pdev_index].sram_size[i];

                if ((extsram_size > 0) && (CPUPhysical >= extsram_base) && (CPUPhysical < (extsram_base + extsram_size))) {
                    *GPUPhysical =  pcie_platform->pcie_info[pdev_index].sram_gpu_base[i] + CPUPhysical - extsram_base;
                    pr_info("axi-sram, from %llx, cpuphysical %llx to gpyphsical %llx.\n", extsram_base, CPUPhysical, *GPUPhysical);
                    return gcvSTATUS_OK;
                }
            }
        }
#endif
    }

    //use default same address
    *GPUPhysical = CPUPhysical;

    return gcvSTATUS_OK;
}

#if gcdENABLE_MP_SWITCH
gceSTATUS
_SwitchCoreCount(gcsPLATFORM *Platform, gctUINT32 *Count)
{
    *Count = Platform->coreCount;

    return gcvSTATUS_OK;
}
#endif

#if gcdENABLE_VIDEO_MEMORY_MIRROR
gceSTATUS
_dmaCopy(gctPOINTER Object, gcsDMA_TRANS_INFO *Info)
{
    gceSTATUS status = gcvSTATUS_OK;
    gckKERNEL kernel = (gckKERNEL)Object;
    gckVIDMEM_NODE dst_node_obj = (gckVIDMEM_NODE)Info->dst_node;
    gckVIDMEM_NODE src_node_obj = (gckVIDMEM_NODE)Info->src_node;
    gctPHYS_ADDR_T src_ptr, dst_ptr;
    gctSIZE_T size0, size1;
    gctSIZE_T srcoffset = 0, dstoffset = 0;
    gctINT ep_fd = 0;

    gcsPLATFORM *platform = kernel->os->device->platform;
    struct _gcsPLATFORM_PCIE *pcie_platform = (struct _gcsPLATFORM_PCIE *)platform;
    struct pci_dev *pdev = pcie_platform->pcie_info[kernel->device->platformIndex].pdev;

    /* Make compiler happy. */
    pdev = pdev;

#if gcdDISABLE_NODE_OFFSET
    Info->offset = 0;
#endif

    status = gckVIDMEM_NODE_GetOffset(kernel, src_node_obj, &srcoffset);
    if (status)
       return status;

    gcmkTRACE_ZONE_N(
                gcvLEVEL_INFO, gcvZONE_ALLOCATOR, 100, "src offset %lx, %lx.\n", srcoffset, Info->offset);
    srcoffset += Info->offset;

    status = gckVIDMEM_NODE_GetOffset(kernel, dst_node_obj, &dstoffset);
    if (status)
        return status;

    gcmkTRACE_ZONE_N(
                gcvLEVEL_INFO, gcvZONE_ALLOCATOR, 100, "dst offset %lx, %lx.\n", dstoffset, Info->offset);
    dstoffset += Info->offset;

    status = gckVIDMEM_NODE_GetSize(kernel, src_node_obj, &size0);
    if (status)
        return status;

    status = gckVIDMEM_NODE_GetSize(kernel, dst_node_obj, &size1);
    if (status)
        return status;

    if (src_node_obj->mirror.type == gcvMIRROR_TYPE_SYSTEM_MEMORY_MIRROR) {
        gctPOINTER tmp_logical = NULL;

        status = gckVIDMEM_NODE_GetGPUPhysical(kernel, src_node_obj, 0, &src_ptr);
        if (status)
            return status;

        status = gckVIDMEM_NODE_LockCPU(kernel, dst_node_obj, 0, 1, &tmp_logical);
        if (status)
            return status;
        dst_ptr = (gctPHYS_ADDR_T)tmp_logical;

        gcmkTRACE_ZONE(gcvLEVEL_INFO, gcvZONE_DRIVER, "EP->RC the cpu logical address is %llx.", dst_ptr);
    } else if (dst_node_obj->mirror.type == gcvMIRROR_TYPE_SYSTEM_MEMORY_MIRROR) {
        gctPOINTER tmp_logical = NULL;

        status = gckVIDMEM_NODE_LockCPU(kernel, src_node_obj, 0, 1, &tmp_logical);
        if (status)
            return status;
        src_ptr = (gctPHYS_ADDR_T)tmp_logical;
        status = gckVIDMEM_NODE_GetGPUPhysical(kernel, dst_node_obj, 0, &dst_ptr);
        if (status)
            return status;
        gcmkTRACE_ZONE(gcvLEVEL_INFO, gcvZONE_DRIVER, "RC->EP the cpu logical address is %llx, %d.", src_ptr);
    }

    //do dma mapping
    if (dst_node_obj->mirror.type == gcvMIRROR_TYPE_SYSTEM_MEMORY_MIRROR) {
        gctPHYS_ADDR_T ep_addr = (gctPHYS_ADDR_T)dst_ptr;

        gcmkTRACE_ZONE_N(
                gcvLEVEL_INFO, gcvZONE_ALLOCATOR, 100, "dst mirror, %llx, %llx.\n", src_ptr, ep_addr);
        status = Xilinx_HdmaTransfer(platform, (gctPHYS_ADDR_T)src_ptr, &ep_addr, gcmMIN(size0, size1), ep_fd, RC2EP);
        dst_ptr = ep_addr;
    } else if (src_node_obj->mirror.type == gcvMIRROR_TYPE_SYSTEM_MEMORY_MIRROR) {
        gctPHYS_ADDR_T ep_addr = (gctPHYS_ADDR_T)src_ptr;

        gcmkTRACE_ZONE_N(
                gcvLEVEL_INFO, gcvZONE_ALLOCATOR, 100, "src mirror. %llx, %llx.\n", src_ptr, ep_addr);
        status = Xilinx_HdmaTransfer(platform, (gctPHYS_ADDR_T)dst_ptr, &ep_addr, gcmMIN(size0, size1), ep_fd, EP2RC);
        src_ptr = src_ptr;
    } else {
        status = -EINVAL;
        dump_stack();
    }

    if (status)
        pr_err("error to do hdma transfer.");

    return status;
}
#endif

#if 0
static int gpu_sub_init(struct pci_dev *pdev, struct _gcsPLATFORM *platform)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 24)
    static u64 dma_mask = DMA_BIT_MASK(40);
#else
    static u64 dma_mask = DMA_40BIT_MASK;
#endif
    struct _gcsPLATFORM_PCIE *pciePlatform = (struct _gcsPLATFORM_PCIE *)platform;

    if (!pciePlatform) {
        pr_err("input param error, platform is empty.\n");
        return -EINVAL;
    }

    gcmkPRINT("PCIE DRIVER PROBED");
    if (pci_enable_device(pdev)) {
        gcmkTRACE_ZONE(gcvLEVEL_ERROR, gcvZONE_EVENT,"galcore: pci_enable_device() failed.\n");
    }

    if (pci_set_dma_mask(pdev, dma_mask)) {
        gcmkTRACE_ZONE(gcvLEVEL_ERROR, gcvZONE_EVENT,"galcore: Failed to set DMA mask.\n");
    }

    pci_set_master(pdev);

    if (pci_request_regions(pdev, "galcore")) {
        gcmkTRACE_ZONE(gcvLEVEL_ERROR, gcvZONE_EVENT,"galcore: Failed to get ownership of BAR region.\n");
    }

#if USE_MSI
    if (pci_enable_msi(pdev)) {
        gcmkTRACE_ZONE(gcvLEVEL_ERROR, gcvZONE_EVENT,"galcore: Failed to enable MSI.\n");
    }
#endif
    return 0;

}

static void gpu_sub_release(struct pci_dev *pdev)
{
    pci_set_drvdata(pdev, NULL);
#if USE_MSI
    pci_disable_msi(pdev);
#endif
    pci_release_regions(pdev);
    return;
}
#endif

int gckPLATFORM_Init(struct pci_dev *pdev, struct _gcsPLATFORM **platform)
{
    int ret = 0;
    struct _gcsPLATFORM_PCIE *pcie_platform = NULL;

    pcie_platform = (struct _gcsPLATFORM_PCIE *)kzalloc(sizeof(struct _gcsPLATFORM_PCIE), GFP_KERNEL);
    if (!pcie_platform) {
        pr_err("galcore: platform application failed.\n");
        return -ENOMEM;
    }

    pcie_platform->base.name = __FILE__;
    pcie_platform->base.ops = &default_ops;
    pcie_platform->pcie_info[pcie_platform->device_number++].pdev = pdev;

    *platform = (struct _gcsPLATFORM *)pcie_platform;

    return ret;
}

int gckPLATFORM_Terminate(struct _gcsPLATFORM *platform)
{
    unsigned int dev_index;
    struct _gcsPLATFORM_PCIE *pcie_platform = (struct _gcsPLATFORM_PCIE *)platform;

    for (dev_index = 0; dev_index < pcie_platform->device_number; dev_index++) {
        unsigned int i;

        for (i = 0; i < gcdMAX_PCIE_BAR; i++) {
            if (pcie_platform->pcie_info[dev_index].bar[i].logical != 0)
                pci_iounmap(pcie_platform->pcie_info[dev_index].pdev,
                            pcie_platform->pcie_info[dev_index].bar[i].logical);
        }
    }

    kfree(pcie_platform);
    pcie_platform = NULL;
    return 0;
}

gceSTATUS Xilinx_HdmaTransfer(
    IN gcsPLATFORM * Platform,
    IN gctADDRESS CPUPhysical,
    IN gctPHYS_ADDR_T *GPUPhysical,
    IN gctSIZE_T size,
    IN gctINT ep_fd,
    IN gctUINT32 dirction
    )
{
    struct trans_pcie_hdma hdma_info;
    struct _gcsPLATFORM_PCIE *pcie_platform = (struct _gcsPLATFORM_PCIE *)Platform;
    struct sn_tranx_t *tdev;

    tdev = (struct sn_tranx_t *)pci_get_drvdata(pcie_platform->pcie_info[0].pdev);
    if (CPUPhysical == 0) {
        gcmkTRACE_ZONE_N(
                         gcvLEVEL_INFO, gcvZONE_ALLOCATOR,
                         40,
                         "warning: CPUPhysical = 0\n"
                        );
        return gcvSTATUS_OK;
    }

    if (dirction == RC2EP) {
        hdma_info.size    = size;
        hdma_info.sar     = CPUPhysical;
        hdma_info.dar     = *GPUPhysical;
        hdma_info.dir     = RC2EP;
        hdma_info.fd      = ep_fd;

#if gcmIS_DEBUG(gcdDEBUG_TRACE)
        gcmkTRACE_ZONE_N(
                gcvLEVEL_INFO, gcvZONE_ALLOCATOR,
                100,
                "RC->EP: 0x%llx ----> 0x%llx  size:0x%x begin\n",
                (gctUINT64)hdma_info.sar, (gctUINT64)hdma_info.dar, hdma_info.size
                );
#endif
        if (hdma_tranx_viraddr_mode(&hdma_info, tdev)) {
            gcmkTRACE_ZONE_N(
                gcvLEVEL_INFO, gcvZONE_ALLOCATOR,
                25,
                "hdma transfer failed\n"
                );
            return gcvSTATUS_INTERRUPTED;
        }

#if gcmIS_DEBUG(gcdDEBUG_TRACE)
        gcmkTRACE_ZONE_N(
                gcvLEVEL_INFO, gcvZONE_ALLOCATOR,
                100,
                "RC->EP: 0x%llx ----> 0x%llx  size:0x%x done\n",
                (gctUINT64)hdma_info.sar, (gctUINT64)hdma_info.dar, hdma_info.size
                );
#endif
    } else if (dirction == EP2RC) {
        hdma_info.size    = size;
        hdma_info.sar     = *GPUPhysical;
        hdma_info.dar     = CPUPhysical;
        hdma_info.dir     = EP2RC;
        hdma_info.fd      = ep_fd;

#if gcmIS_DEBUG(gcdDEBUG_TRACE)
        gcmkTRACE_ZONE_N(
                gcvLEVEL_INFO, gcvZONE_ALLOCATOR,
                100,
                "EP->RC: 0x%llx ----> 0x%llx  size:0x%x begin\n",
                (gctUINT64)hdma_info.sar, (gctUINT64)hdma_info.dar, hdma_info.size
                );
#endif

        if (hdma_tranx_viraddr_mode(&hdma_info, tdev)) {
            gcmkTRACE_ZONE_N(
                gcvLEVEL_INFO, gcvZONE_ALLOCATOR,
                25,
                "hdma transfer failed\n"
                );
            return gcvSTATUS_INTERRUPTED;
        }

#if gcmIS_DEBUG(gcdDEBUG_TRACE)
        gcmkTRACE_ZONE_N(
                gcvLEVEL_INFO, gcvZONE_ALLOCATOR,
                100,
                "EP->RC: 0x%llx ----> 0x%llx  size:0x%x done\n",
                (gctUINT64)hdma_info.sar, (gctUINT64)hdma_info.dar, hdma_info.size
                );
#endif
    } else {
            gcmkTRACE_ZONE_N(
                gcvLEVEL_INFO, gcvZONE_ALLOCATOR,
                25,
                "wrong dma direction\n"
                );

    }
    return gcvSTATUS_OK;
}

